from __future__ import absolute_import
import chardet
import copy
import json
import imp
import multiprocessing
import os
import re
import signal
import sys
import time
import warnings
import zipcodes

try:
    from importlib.machinery import SourceFileLoader
except ImportError:
    import imp

import pandas as pd
import numpy as np

from contextlib import closing
from hashlib import md5
from random import uniform
from scandir import walk
from unidecode import unidecode


class classproperty(object):

    """
    This decorator allows you to define functions on a class that are accessible directly from the
    class itself (rather than an instance of the class). It allows you to access ``classproperty``
    attributes directly, such as ``obj.property``, rather than as a function on a class instance
    (like ``obj = Obj(); obj.property()``).

    Borrowed from a StackOverflow `post <https://stackoverflow.com/a/3203659>`_.

    Usage::

        from pewtils import classproperty

        class MyClass(object):
            x = 4

            @classproperty
            def number(cls):
                return cls.x

        >>> MyClass().number
        4
        >>> MyClass.number
        4
    """

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


def is_not_null(val, empty_lists_are_null=False, custom_nulls=None):

    """
    Checks whether the value is null, using a variety of potential string values, etc. The following values are always
    considered null: ``numpy.nan, None, "None", "nan", "", " ", "NaN", "none", "n/a", "NONE", "N/A"``

    :param val: The value to check
    :param empty_lists_are_null: Whether or not an empty list or :py:class:`pandas.DataFrame` should be considered \
    null (default=False)
    :type empty_lists_are_null: bool
    :param custom_nulls: an optional list of additional values to consider as null
    :type custom_nulls: list
    :return: True if the value is not null
    :rtype: bool

    Usage::

        from pewtils import is_not_null

        >>> text = "Hello"
        >>> is_not_null(text)
        True
    """

    null_values = [None, "None", "nan", "", " ", "NaN", "none", "n/a", "NONE", "N/A"]
    if custom_nulls:
        null_values.extend(custom_nulls)
    if type(val) == list:
        if empty_lists_are_null and val == []:
            return False
        else:
            return True
    elif isinstance(val, pd.Series) or isinstance(val, pd.DataFrame):
        if empty_lists_are_null and len(val) == 0:
            return False
        else:
            return True
    else:
        try:
            try:
                good = val not in null_values
                if good:
                    try:
                        try:
                            good = not pd.isnull(val)
                        except IndexError:
                            good = True
                    except AttributeError:
                        good = True
                return good
            except ValueError:
                return val.any()
        except TypeError:
            return not isinstance(val, None)


def is_null(val, empty_lists_are_null=False, custom_nulls=None):

    """
    Returns the opposite of the outcome of :py:func:`pewtils.is_not_null`. The following values are always \
    considered null: ``numpy.nan, None, "None", "nan", "", " ", "NaN", "none", "n/a", "NONE", "N/A"``

    :param val: The value to check
    :param empty_lists_are_null: Whether or not an empty list or :py:class:`pandas.DataFrame` should be considered \
    null (default=False)
    :type empty_lists_are_null: bool
    :param custom_nulls: an optional list of additional values to consider as null
    :type custom_nulls: list
    :return: True if the value is null
    :rtype: bool

    Usage::

        from pewtils import is_null

        >>> empty_list = []
        >>> is_null(empty_list, empty_lists_are_null=True)
        True
    """

    return not is_not_null(
        val, empty_lists_are_null=empty_lists_are_null, custom_nulls=custom_nulls
    )


def decode_text(text, throw_loud_fail=False):

    """
    Attempts to decode and re-encode text as ASCII. In the case of failure, it will attempt to detect the string's \
    encoding, decode it, and convert it to ASCII. If both these attempts fail, it will attempt to use the \
    :py:mod:`unidecode` package to transliterate into ASCII. And finally, if that doesn't work, it will forcibly \
    encode the text as ASCII and ignore non-ASCII characters.

    .. warning:: This function is potentially destructive to source input and should be used with some care. \
        Input text that cannot be decoded may be stripped out, or replaced with a similar ASCII character or other \
        placeholder, potentially resulting in an empty string.

    :param text: The text to process
    :type text: str
    :param throw_loud_fail: If True, exceptions will be raised, otherwise the function will fail silently and \
    return an empty string (default False)
    :type throw_loud_fail: bool
    :return: Decoded text, or empty string
    :rtype: str

    .. note:: In Python 3, the decode/encode attempts will fail by default, and the :py:mod:`unidecode` package will \
        be used to transliterate. In general, you shouldn't need to use this function in Python 3, but it shouldn't \
        hurt anything if you do.

    """

    output_text = ""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if is_not_null(text):
            try:
                text = u"{}".format(text)
                output_text = text.decode("ascii").encode("ascii")
            except (AttributeError, TypeError, UnicodeEncodeError, UnicodeDecodeError):
                try:
                    output_text = text.decode(chardet.detect(text)["encoding"])
                    output_text = output_text.encode("ascii")
                except (
                    AttributeError,
                    TypeError,
                    UnicodeEncodeError,
                    UnicodeDecodeError,
                ):
                    try:
                        output_text = unidecode(text)
                    except (
                        AttributeError,
                        TypeError,
                        UnicodeEncodeError,
                        UnicodeDecodeError,
                    ):
                        if throw_loud_fail:
                            output_text = text.decode("ascii", "ignore").encode("ascii")
                        else:
                            try:
                                output_text = text.decode("ascii", "ignore").encode(
                                    "ascii"
                                )
                            except (
                                AttributeError,
                                TypeError,
                                UnicodeEncodeError,
                                UnicodeDecodeError,
                            ):
                                print("Could not decode")
                                print(text)
                output_text = output_text.replace("\x00", "").replace("\u0000", "")

    return output_text


def get_hash(text, hash_function="ssdeep"):

    """
    Generates hashed text using one of several available hashing functions.

    :param text: The string to hash
    :type text: str
    :param hash_function: The specific algorithm to use; options are ``'nilsimsa'``, ``'md5'``, and ``'ssdeep'`` \
    (default)
    :type hash_function: str
    :return: A hashed representation of the provided string
    :rtype: str

    .. note:: The string will be passed through :py:func:`pewtils.decode_text` and the returned value will be used \
    instead of the original value if it runs successfully, in order to ensure consistent hashing in both Python 2 and \
    3. By default the function uses the :py:mod:`ssdeep` algorithm, which generates context-sensitive hashes that are \
    useful for computing document similarities at scale.

    .. note:: Using `hash_function='ssdeep'` requires the :py:mod:`ssdeep` library, which is not installed by default \
    because it requires the installation of additional system libraries on certain operating systems. For help \
    installing ssdeep, refer to the pewtils documentation installation section, which provides OS-specific instructions.

    Usage::

        from pewtils import get_hash

        >>> text = 'test_string'
        >>> get_hash(text)
        '3:HI2:Hl'
    """

    decoded_text = decode_text(text).encode("utf8").strip()
    if decoded_text == "":
        decoded_text = text
    text = decoded_text
    if hash_function == "nilsimsa":
        from nilsimsa import Nilsimsa

        hashed = Nilsimsa(text).hexdigest()
    elif hash_function == "md5":
        hashed = md5(text).hexdigest()
    else:
        try:
            import ssdeep
        except ImportError:
            raise Exception(
                """
                To use get_hash with hash_function='ssdeep' you need to install the ssdeep package. Try running: 
                    >> BUILD_LIB=1 pip install ssdeep
                If you encounter installation problems, refer to the pewtils documentation for troubleshooting help.
            """
            )
        hashed = ssdeep.hash(text)

    return hashed


def zipcode_num_to_string(zipcode):

    """
    Attempts to standardize a string/integer/float that contains a U.S. zipcode. Front-pads with zeroes and uses the \
    :py:mod:`zipcodes` library to ensure that the zipcode is real. If the zipcode doesn't validate successfully, \
    ``None`` will be returned.

    :param zip: Object that contains a sequence of digits (string, integer, float)
    :type zip: str or float or int
    :return: A 5-digit string, or None
    :rtype: str or NoneType

    Usage::

        from pewtils import zipcode_num_to_string

        >>> zipcode_number = 6463
        >>> zipcode_num_to_string(zipcode_number)
        '06463'
        >>> not_zipcode_number = 345678
        >>> zipcode_num_to_string(not_zipcode_number)
        >>>
    """

    if is_not_null(zipcode):

        try:
            zipcode = str(int(str(zipcode).strip()[:5].split(".")[0]))
        except (TypeError, ValueError):
            zipcode = None

        if zipcode:
            zipcode = zipcode.zfill(5)
            if zipcodes.is_real(zipcode):
                return zipcode
            else:
                return None
    else:

        zipcode = None

    return zipcode


def concat_text(*args):

    """
    A helper function for concatenating text values. Text values are passed through :py:func:`pewtils.decode_text` \
    before concatenation.

    :param args: A list of text values that will be returned as a single space-separated string
    :type args: list
    :return: A single string of the values concatenated by spaces
    :rtype: str

    Usage::

        from pewtils import concat_text

        >>> text_list = ['Hello', 'World', '!']
        >>> concat_text(text_list)
        'Hello World !'
    """

    strs = [decode_text(arg) for arg in args if is_not_null(arg)]
    return " ".join(strs) if is_not_null(strs, empty_lists_are_null=True) else ""


def vector_concat_text(*args):

    """
    Takes a list of equal-length lists and returns a single list with the rows concatenated by spaces. Useful for \
    merging multiple columns of text in Pandas.

    :param args: A list of lists or :py:class:`pandas.Series` s that contain text values
    :return: A single list or :py:class:`pandas.Series` with all of the text values for each row concatenated

    Usage with lists::

        from pewtils import vector_concat_text

        >>> text_lists = ["one", "two", "three"], ["a", "b", "c"]
        >>> vector_concat_text(text_lists)
        ['one a', 'two b', 'three c']

    Usage with Pandas::

        import pandas as pd
        from pewtils import vector_concat_text

        df = pd.DataFrame([
            {"text1": "one", "text2": "a"},
            {"text1": "two", "text2": "b"},
            {"text1": "three", "text2": "c"}
        ])

        >>> df['text'] = vector_concat_text(df['text1'], df['text2'])
        >>> df['text']
        0      one a
        1      two b
        2    three c
        Name: text, dtype: object
    """

    return np.vectorize(concat_text)(*args)


def scale_range(old_val, old_min, old_max, new_min, new_max):

    """
    Scales a value from one range to another.  Useful for comparing values from different scales, for example.

    :param old_val: The value to convert
    :type old_val: int or float
    :param old_min: The minimum of the old range
    :type old_min: int or float
    :param old_max: The maximum of the old range
    :type old_max: int or float
    :param new_min: The minimum of the new range
    :type new_min: int or float
    :param new_max: The maximum of the new range
    :type new_max: int or float
    :return: Value equivalent from the new scale
    :rtype: float

    Usage::

        from pewtils import scale_range

        >>> old_value = 5
        >>> scale_range(old_value, 0, 10, 0, 20)
        10.0
    """

    return (
        ((float(old_val) - float(old_min)) * (float(new_max) - float(new_min)))
        / (float(old_max) - float(old_min))
    ) + float(new_min)


def new_random_number(attempt=1, minimum=1.0, maximum=10):

    """
    Returns a random number between the boundary that exponentially increases with the number of ``attempt``.
    The upper bound is capped using the ``maximum`` parameter (default 10) but is otherwise determined by the
    function ``minimum * 2 ** attempt``.

    | In effect, this means that when ``attempt`` is 1, the number returned will be in the range of the minimum \
    and twice the minimum's value.  As you increase ``attempt``, the possible range of returned values expands \
    exponentially until it hits the ``maximum`` ceiling.

    :param attempt: Increasing attempt will expand the upper-bound of the range from which the random number is drawn
    :type attempt: int
    :param minimum: The minimum allowed value that can be returned; must be greater than zero.
    :type minimum: int or float
    :param maximum: The maximum allowed value that can be returned; must be greater than ``minimum``.
    :type maximum: int or float
    :return: A random number drawn uniformly from across the range determined by the provided arguments.
    :rtype: float

    .. note:: One useful application of this function is rate limiting: a script can pause in between requests at a \
        reasonably fast pace, but then moderate itself and pause for longer periods if it begins encountering errors, \
        simply by increasing the ``attempt`` variable (hence its name).

    Usage::

        from pewtils import new_random_number

        >>> new_random_number(attempt=1)
        1.9835581813820642
        >>> new_random_number(attempt=2)
        3.1022350739064
    """

    return uniform(minimum, min(maximum, minimum * 2 ** attempt))


def chunk_list(seq, size):

    """
    Takes a sequence and groups values into smaller lists based on the specified size.

    :param seq: List or a list-like iterable
    :type seq: list or iterable
    :param size: Desired size of each sublist
    :type size: int
    :return: A list of lists
    :rtype: list

    Usage::

        from pewtils import chunk_list

        >>> number_sequence = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        >>> chunk_list(number_sequence, 3)
        [[1, 2, 3], [4, 5, 6], [7, 8, 9], [10]]
    """

    return (seq[pos : (pos + size)] for pos in range(0, len(seq), size))


def flatten_list(l):

    """
    Takes a list of lists and flattens it into a single list. Nice shortcut to avoid having to deal with list \
    comprehension.

    :param l: A list of lists
    :type l: list
    :return: A flattened list of all of the elements contained in the original list of lists
    :rtype: list

    Usage::

        from pewtils import flatten_list

        >>> nested_lists = [[1, 2, 3], [4, 5, 6]]
        >>> flatten_list(nested_lists)
        [1, 2, 3, 4, 5, 6]
    """

    return [item for sublist in l for item in sublist]


def scan_dictionary(search_dict, field):

    """
    Takes a dictionary with nested lists and dictionaries, and searches recursively for a specific key. Since keys can
    occur more than once, the function returns a list of all of the found values along with a list of equal length
    that specifies the nested key path to each value.

    :param search_dict: The dictionary to search
    :type search_dict: dict
    :param field: The field to find
    :type field: str
    :return: A tuple of the found values and file path-style strings representing their locations
    :rtype: tuple

    Usage::

        from pewtils import scan_dictionary

        >>> test_dict = {"one": {"two": {"three": "four"}}}
        >>> scan_dictionary(test_dict, "three")
        (['four'], ['one/two/three/'])
        >>> scan_dictionary(test_dict, "five")
        ([], [])
    """

    fields_found = []
    key_path = []

    for key, value in search_dict.items():
        if key == field:
            fields_found.append(value)
            new_str = str(key) + "/"
            key_path.append(new_str)

        elif isinstance(value, dict):
            results, path = scan_dictionary(value, field)
            for result in results:
                fields_found.append(result)
            for road in path:
                new_str = str(key) + "/" + road
                key_path.append(new_str)

        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    more_results, more_path = scan_dictionary(item, field)
                    for another_result in more_results:
                        fields_found.append(another_result)
                    for another_road in more_path:
                        new_str = str(key) + "/" + another_road
                        key_path.append(new_str)

    return fields_found, key_path


def recursive_update(existing, new):

    """
    Takes an object and a dictionary representation of attributes and values, and recursively traverses through the
    new values and updates the object.

    | Regardless of whether or not the keys in the dictionary correspond to attribute names or dictionary keys; \
    you can use this to iterate through a nested hierarchy of objects and dictionaries and update whatever you like.

    :param existing: An object or dictionary
    :type existing: dict or object
    :param new: A dictionary where keys correspond to the names of keys in the existing dictionary or attributes on \
    the existing object
    :type new: dict or object
    :return: A copy of the original object or dictionary, with the values updated based on the provided map
    :rtype: dict or object

    Usage::

        from pewtils import recursive_update

        class TestObject(object):
            def __init__(self, value):
                self.value = value
                self.dict = {"obj_key": "original"}
            def __repr__(self):
                return("TestObject(value='{}', dict={})".format(self.value, self.dict))

        original = {
            "object": TestObject("original"),
            "key1": {"key2": "original"}
        }
        update = {
            "object": {"value": "updated", "dict": {"obj_key": "updated"}},
            "key1": {"key3": "new"}
        }

        >>> recursive_update(original, update)
        {'object': TestObject(value='updated', dict={'obj_key': 'updated'}),
         'key1': {'key2': 'original', 'key3': 'new'}}

    """

    def _hasattr(obj, attr):
        if isinstance(obj, dict):
            return attr in obj
        else:
            return hasattr(obj, attr)

    def _setattr(obj, attr, val):
        if isinstance(obj, dict):
            obj[attr] = val
        else:
            setattr(obj, attr, val)
        return obj

    def _getattr(obj, attr):
        if isinstance(obj, dict):
            return obj[attr]
        else:
            return getattr(obj, attr)

    existing = copy.deepcopy(existing)
    if isinstance(new, dict):
        for k, v in new.items():

            if _hasattr(existing, k):
                _setattr(
                    existing,
                    k,
                    recursive_update(_getattr(existing, k), _getattr(new, k)),
                )
            else:
                _setattr(existing, k, _getattr(new, k))
        return existing
    else:
        return new


def cached_series_mapper(series, function):

    """
    Applies a function to all of the unique values in a :py:class:`pandas.Series` to avoid repeating the operation \
    on duplicate values.

    | Great if you're doing database lookups or something computationally intensive on a column that may contain \
    repeating values, etc.

    :param series: A :py:class:`pandas.Series`
    :type series: :py:class:`pandas.Series`
    :param function: A function to apply to values in the :py:class:`pandas.Series`
    :return: The resulting :py:class:`pandas.Series`
    :rtype: :py:class:`pandas.Series`

    Usage::

        import pandas as pd
        from pewtils import cached_series_mapper

        values = ["value"]*10
        def my_function(x):
            print(x)
            return x

        df = pd.DataFrame(values, columns=['column'])
        >>> mapped = df['column'].map(my_function)
        value
        value
        value
        value
        value
        value
        value
        value
        value
        value
        >>> mapped = cached_series_mapper(df['column'], my_function)
        value
    """

    val_map = {}
    for val in series.unique():
        val_map[val] = function(val)

    return series.map(val_map)


def multiprocess_group_apply(grp, func, *args, **kwargs):
    """

    Apply arbitrary functions to groups or slices of a Pandas DataFrame using multiprocessing, to efficiently \
    map or aggregate data. Each group gets processed in parallel, and the results are concatenated together after \
    all processing has finished. If you pass a function that aggregates each group into a single value, you'll get \
    back a DataFrame with one row for each group, as though you had performed a `.agg` function. \
    If you pass a function that returns a value for each _row_ in the group, then you'll get back a DataFrame \
    in your original shape. In this case, you would simply be using grouping to efficiently apply a row-level operation.

    :param grp: A Pandas DataFrameGroupBy object
    :type grp: pandas.core.groupby.generic.DataFrameGroupBy
    :param func: A function that accepts a Pandas DataFrame representing a group from the original DataFrame
    :type func: function
    :param args: Arguments to be passed to the function
    :param kwargs: Keyword arguments to be passed to the function
    :return: The resulting DataFrame
    :rtype: pandas.DataFrame

    Usage::

        df = pd.DataFrame([
            {"group": 1, "value": "one two three"},
            {"group": 1, "value": "one two three four"},
            {"group": 2, "value": "one two"}
        ])

        ### For efficient aggregation

        def get_length(grp):
            # Simple function that returns the number of rows in each group
            return len(grp)

        >>> df.groupby("group_col").apply(lambda x: len(x))
        1    2
        2    1
        dtype: int64
        >>> multiprocess_group_apply(df.groupby("group_col"), get_length)
        1    2
        2    1
        dtype: int64

        ### For efficient mapping

        def get_value_length(grp):
            # Simple function that returns the word count of each row in the group
            return grp['value'].map(lambda x: len(x.split()))

        >>> df['value'].map(lambda x: len(x.split()))
        0    3
        1    4
        2    2
        Name: value, dtype: int64
        >>> multiprocess_group_apply(df.groupby("group_col"), get_value_length)
        0    3
        1    4
        2    2
        Name: value, dtype: int64

        # If you just want to efficiently map a function to your DataFrame and you want to evenly split your
        # DataFrame into groups, you could do the following:

        df["group_col"] = (df.reset_index().index.values / (len(df) / multiprocessing.cpu_count())).astype(int)
        df["mapped_value"] = multiprocess_group_apply(df.groupby("group_col"), get_value_length)
        del df["group_col"]

    """

    results = []
    pool = multiprocessing.Pool(processes=multiprocessing.cpu_count())
    for name, group in grp:
        results.append(pool.apply_async(func, (group,) + args, kwargs))
    pool.close()
    pool.join()
    results = [r.get() for r in results]

    if not hasattr(results[0], "__len__") or isinstance(results[0], str):
        # Assume it's an aggregation function
        return pd.Series(results, index=[g for g, _ in grp])
    else:
        # Assume you're just mapping the function normally and using the groups to split the data
        return pd.concat(results)


def extract_json_from_folder(
    folder_path, include_subdirs=False, concat_subdir_names=False
):

    """
    Takes a folder path and traverses it, looking for JSON files. When it finds one, it adds it to a dictionary,
    with the key being the name of the file and the value being the JSON itself. This is useful if you store \
    configurations or various metadata in a nested folder structure, which we do for things like content analysis \
    codebooks.

    | Has options for recursively traversing a folder, and for optionally concatenating the subfolder names \
    into the dictionary keys as prefixes.

    :param folder_path: The path of the folder to scan
    :type folder_path: str
    :param include_subdirs: Whether or not to recursively scan subfolders
    :type include_subdirs: bool
    :param concat_subdir_names: Whether or not to prefix the dictionary keys with the names of subfolders
    :type concat_subdir_names: bool
    :return: A dictionary containing all of the abstracted JSON files as values
    :rtype: dict

    Usage::

        # For example, let's say we have the following folder structure
        # with various JSON codebooks scattered about:
        #
        # /codebooks
        #     /logos
        #         /antipathy.json
        #     /atp_open_ends
        #         /w29
        #             /sources_of_meaning.json
        #
        # Here's what we'd get depending on the different parameters we use:

        from pewtils import extract_json_from_folder
        >>> extract_json_from_folder("codebooks", include_subdirs=False, concat_subdir_names=False)
        {}
        >>> extract_json_from_folder("codebooks", include_subdirs=True, concat_subdir_names=False)
        {
            "logos": {"antipathy": "json would be here"},
            "atp_open_ends": {"w29": {"sources_of_meaning": "json would be here"}}
        }
        >>> extract_json_from_folder("codebooks", include_subdirs=True, concat_subdir_names=True)
        {
            "logos_antipathy": "json would be here",
            "atp_open_ends_w29_sources_of_meaning": "json would be here"
        }
    """

    attributes = {}
    subdirs = []
    if os.path.exists(folder_path):
        for path, subdir, files in walk(folder_path):
            if folder_path == path:
                for file in files:
                    if file.endswith(".json"):
                        key = re.sub(".json", "", file)
                        with closing(open(os.path.join(path, file), "r")) as infile:
                            try:
                                attributes[key] = json.load(infile)
                            except ValueError:
                                print("JSON file is invalid: {}".format(file))
            if subdir:
                subdirs.append(subdir)

    if include_subdirs and len(subdirs) > 0:
        for subdir in subdirs[0]:
            if subdir != "__pycache__":
                results = extract_json_from_folder(
                    os.path.join(folder_path, subdir),
                    include_subdirs=True,
                    concat_subdir_names=concat_subdir_names,
                )
                if not concat_subdir_names:
                    attributes[subdir] = results
                else:
                    for subattr_name, subattr in results.items():
                        attributes["_".join([subdir, subattr_name])] = subattr

    return attributes


def extract_attributes_from_folder_modules(
    folder_path,
    attribute_name,
    include_subdirs=False,
    concat_subdir_names=False,
    current_subdirs=None,
):

    """
    Takes a folder path and traverses it, looking for Python files that contain an attribute (i.e., class, function,
    etc.) with a given name. It extracts those attributes and returns a dictionary where the keys are the names of the
    files that contained the attributes, and the values are the attributes themselves.

    This operates exactly the same as :py:func:`pewtils.extract_json_from_folder` except instead of reading JSON files
    and adding them as values in the dictionary that gets returned, this function will instead look for Python files
    that contain a function, class, method, or attribute with the name you provide in ``attribute_name`` and will load
    that attribute in as the values.

    :param folder_path: The path of a folder/module to scan
    :type folder_path: str
    :param attribute_name: The name of the attribute (class, function, variable, etc.) to extract from files
    :type attribute_name: str
    :param include_subdirs: Whether or not to recursively scan subfolders
    :type include_subdirs: bool
    :param concat_subdir_names: Whether or not to prefix the dictionary keys with the names of subfolders
    :type concat_subdir_names: bool
    :param current_subdirs: Used to track location when recursively iterating a module (do not use)
    :return: A dictionary with all of the extracted attributes as values
    :rtype: dict

    .. note:: if you use Python 2.7 you will need to add ``from __future__ import absolute_import`` to the top of files \
        that you want to scan and import using this function.
    """

    if not folder_path.startswith(os.getcwd()):
        folder_path = os.path.join(os.getcwd(), folder_path)
    test_path, _ = os.path.split(folder_path)
    while test_path != "/":
        if "__init__.py" not in os.listdir(test_path):
            break
        test_path, _ = os.path.split(test_path)
    module_location = test_path

    current_folder = folder_path.split("/")[-1]
    if not current_subdirs:
        current_subdirs = []

    attributes = {}
    subdirs = []
    if os.path.exists(folder_path):
        for path, subdir_list, files in walk(folder_path):
            if folder_path == path:
                for file in files:
                    if file.endswith(".py") and not file.startswith("__init__"):
                        file_name = file.split(".")[0]
                        module_name = re.sub(
                            "/",
                            ".",
                            re.sub(
                                module_location,
                                "",
                                os.path.splitext(os.path.join(path, file))[0],
                            ),
                        ).strip(".")
                        if module_name in sys.modules:
                            module = sys.modules[module_name]
                            # https://github.com/ansible/ansible/issues/13110
                        else:
                            try:
                                module = SourceFileLoader(
                                    module_name, os.path.join(path, file)
                                ).load_module()
                            except NameError:
                                file, pathname, description = imp.find_module(
                                    file_name, [path]
                                )
                                warnings.simplefilter("error", RuntimeWarning)
                                try:
                                    module = imp.load_module(
                                        module_name, file, pathname, description
                                    )
                                except RuntimeWarning:
                                    try:
                                        module = imp.load_module(
                                            module_name.split(".")[-1],
                                            file,
                                            pathname,
                                            description,
                                        )
                                    except RuntimeWarning:
                                        module = None
                                    except (ImportError, AttributeError):
                                        module = None
                                except (ImportError, AttributeError):
                                    module = None
                        if hasattr(module, attribute_name):
                            attributes[file_name] = getattr(module, attribute_name)

            if subdir_list:
                subdirs.extend(subdir_list)

    if include_subdirs:
        for subdir in set(subdirs):
            results = extract_attributes_from_folder_modules(
                os.path.join(folder_path, subdir),
                attribute_name,
                concat_subdir_names=concat_subdir_names,
                include_subdirs=True,
                current_subdirs=current_subdirs + [current_folder],
            )
            if not concat_subdir_names:
                attributes[subdir] = results
            else:
                for subattr_name, subattr in results.items():
                    attributes["_".join([subdir, subattr_name])] = subattr

    if is_null(current_subdirs, empty_lists_are_null=True):
        for name in attributes.keys():
            try:
                attributes[name]._name = name
            except AttributeError:
                pass

    return attributes


class timeout_wrapper:
    def __init__(self, seconds=1, error_message="Timeout"):
        """
        Context manager that will raise an error if it takes longer than the specified number of seconds to execute.
        Found via this very helpful Stack Overflow post:
        https://stackoverflow.com/questions/2281850/timeout-function-if-it-takes-too-long-to-finish

        :param seconds: Number of seconds allowed for the code to execute
        :param error_message: Optional custom error message to raise
        """
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise Exception(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, t, value, traceback):
        signal.alarm(0)


class PrintExecutionTime(object):

    """
    Simple context manager to print the time it takes for a block of code to execute

    :param label: A label to print alongside the execution time
    :param stdout: a StringIO-like output stream (sys.stdout by default)

    Usage::

        from pewtils import PrintExecutionTime

        >>> with PrintExecutionTime(label="my function"): time.sleep(5)
        my function: 5.004292011260986 seconds

    """

    def __init__(self, label=None, stdout=None):
        self.start_time = None
        self.end_time = None
        self.label = label
        self.stdout = sys.stdout if not stdout else stdout

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.end_time = time.time()
        if self.label:
            self.stdout.write(
                "{}: {} seconds".format(self.label, self.end_time - self.start_time)
            )
        else:
            self.stdout.write("{} seconds".format(self.end_time - self.start_time))



from __future__ import division
from bs4 import BeautifulSoup
from builtins import str
from pewtils import get_hash, decode_text, is_not_null
from six.moves.urllib import parse as urlparse
from unidecode import unidecode
import pandas as pd
import re
import os
import requests
import tldextract
import warnings
from requests.exceptions import ReadTimeout
from stopit import ThreadingTimeout as Timeout


_ = pd.read_csv(
    os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "general_link_shorteners.csv"
    )
)
GENERAL_LINK_SHORTENERS = _["shortener"].values


_ = pd.read_csv(
    os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "vanity_link_shorteners.csv"
    )
)
_ = _[_["historical"] == 0]
VANITY_LINK_SHORTENERS = dict(zip(_["shortener"], _["expanded"]))

_ = pd.read_csv(
    os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "vanity_link_shorteners.csv"
    )
)
_ = _[_["historical"] == 1]
HISTORICAL_VANITY_LINK_SHORTENERS = dict(zip(_["shortener"], _["expanded"]))

VANITY_LINK_SHORTENERS.update(HISTORICAL_VANITY_LINK_SHORTENERS)


def hash_url(url):

    """
    Clears out http/https prefix and returns an MD5 hash of the URL. More effective \
    when used in conjunction with :py:func:`pewtils.http.canonical_link`.

    :param url: The URL to hash
    :type url: str
    :return: Hashed string representation of the URL using the md5 hashing algorithm.
    :rtype: str

    Usage::

        from pewtils.http import hash_url

        >>> hash_url("http://www.example.com")
        "7c1767b30512b6003fd3c2e618a86522"
        >>> hash_url("www.example.com")
        "7c1767b30512b6003fd3c2e618a86522"

    """

    http_regex = re.compile(r"^http(s)?\:\/\/")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        result = get_hash(
            unidecode(http_regex.sub("", url.lower())), hash_function="md5"
        )
        return result


def strip_html(html, simple=False, break_tags=None):

    """
    Attempts to strip out HTML code from an arbitrary string while preserving meaningful text components. \
    By default, the function will use BeautifulSoup to parse the HTML. Setting ``simple=True`` will make the \
    function use a much simpler regular expression approach to parsing.

    :param html: The HTML to process
    :type html: str
    :param simple: Whether or not to use a simple regex or more complex parsing rules (default=False)
    :type simple: bool
    :param break_tags: A custom list of tags on which to break (default is ["strong", "em", "i", "b", "p"])
    :type break_tags: list
    :return: The text with HTML components removed
    :rtype: str

    .. note: This function might not be effective for *all* variations of HTML structures, but it produces fairly \
        reliable results in removing the vast majority of HTML without stripping out valuable content.

    Usage::

        from pewtils.http import strip_html

        >>> my_html = "<html><head>Header text</head><body>Body text</body></html>"
        >>> strip_html(my_html)
        'Header text Body text'

    """

    html = re.sub(r"\n", " ", html)
    html = re.sub(r"\s+", " ", html)
    if not break_tags:
        break_tags = ["strong", "em", "i", "b", "p"]
    if not simple:
        try:

            split_re = re.compile(r"\s{2,}")
            soup = BeautifulSoup(html, "lxml")
            for tag in soup():
                if (
                    "class" in tag.attrs
                    and ("menu" in tag.attrs["class"] or "header" in tag.attrs["class"])
                ) or ("menu" in str(tag.id) or "header" in str(tag.id)):
                    tag.extract()
            for tag in soup(["script", "style"]):
                tag.extract()
            for br in soup.find_all("br"):
                br.replace_with("\n")
            for t in soup(break_tags):
                try:
                    t.replace_with("\n{0}\n".format(t.text))
                except (UnicodeDecodeError, UnicodeEncodeError):
                    t.replace_with("\n{0}\n".format(decode_text(t.text)))
            if hasattr(soup, "body") and soup.body:
                text = soup.body.get_text()
            else:
                text = soup.get_text()
            lines = [l.strip() for l in text.splitlines()]
            lines = [l2.strip() for l in lines for l2 in split_re.split(l)]
            text = "\n".join([l for l in lines if l])
            text = re.sub(r"(\sA){2,}\s", " ", text)
            text = re.sub(r"\n+(\s+)?", "\n\n", text)
            text = re.sub(r" +", " ", text)
            text = re.sub(r"\t+", " ", text)

            return text

        except Exception as e:

            print("strip_html error")
            print(e)
            text = re.sub(r"<[^>]*>", " ", re.sub("\\s+", " ", html)).strip()
            return text

    else:
        return "\n".join(
            [
                re.sub(r"\s+", " ", re.sub(r"\<[^\>]+\>", " ", section))
                for section in re.sub(r"\<\/?div\>|\<\/?p\>|\<br\>", "\n", html).split(
                    "\n"
                )
            ]
        )


def trim_get_parameters(url, session=None, timeout=30, user_agent=None):

    """
    Takes a URL (presumed to be the final end point) and iterates over GET parameters, attempting to find optional
    ones that can be removed without generating any redirects.

    :param url: The URL to trim
    :type url: str
    :param session: (Optional) A persistent session that can optionally be passed (useful if you're processing many \
    links at once)
    :type session: :py:class:`requests.Session` object
    :param user_agent: User agent for the auto-created requests Session to use, if a preconfigured requests Session \
    is not provided
    :type user_agent: str
    :param timeout: Timeout for requests
    :type timeout: int or float
    :return: The original URL with optional GET parameters removed
    :rtype: str

    Usage::

        from pewtils.http import trim_get_parameters

        >>> trim_get_parameters("https://httpbin.org/status/200?param=1")
        "https://httpbin.org/status/200"

    """

    close_session = False
    if not session:
        close_session = True
        session = requests.Session()
        session.headers.update({"User-Agent": user_agent})

    # Often there's extra information about social sharing and referral sources that can be removed
    ditch_params = []
    parsed = urlparse.urlparse(url)
    if parsed.query:
        params = urlparse.parse_qs(parsed.query)
        for k, v in params.items():
            # We iterate over all of the GET parameters and try holding each one out
            check = True
            for skipper in ["document", "article", "id", "qs"]:
                # If the parameter is named something that's probably a unique ID, we'll keep it
                if skipper in k.lower():
                    check = False
            for skipper in ["html", "http"]:
                # Same goes for parameters that contain URL information
                if skipper in v[0].lower():
                    check = False
            if check:
                new_params = {
                    k2: v2[0] for k2, v2 in params.items() if k2 != k and len(v2) == 1
                }
                new_params = urlparse.urlencode(new_params)
                new_parsed = parsed._replace(query=new_params)
                new_url = urlparse.urlunparse(new_parsed)
                try:
                    resp = session.head(new_url, allow_redirects=True, timeout=timeout)
                except ReadTimeout:
                    resp = None
                if is_not_null(resp):
                    new_parsed = urlparse.urlparse(resp.url)
                    if new_parsed.query != "" or new_parsed.path not in ["", "/"]:
                        # If removing a parameter didn't redirect to a root domain...
                        new_url = resp.url
                        compare_new = (
                            new_url.split("?")[0] if "?" in new_url else new_url
                        )
                        compare_old = url.split("?")[0] if "?" in url else url
                        if compare_new == compare_old:
                            # And the domain is the same as it was before, then the parameter was probably unnecessary
                            ditch_params.append(k)

    if len(ditch_params) > 0:
        # Now we remove all of the unnecessary get parameters and finalize the URL
        new_params = {
            k: v[0] for k, v in params.items() if len(v) == 1 and k not in ditch_params
        }
        new_params = urlparse.urlencode(new_params)
        parsed = parsed._replace(query=new_params)
        url = urlparse.urlunparse(parsed)

    if close_session:
        session.close()

    return url


def extract_domain_from_url(
    url,
    include_subdomain=True,
    resolve_url=False,
    timeout=1.0,
    session=None,
    user_agent=None,
    expand_shorteners=True,
):

    """
    Attempts to extract a standardized domain from a url by following the link and extracting the TLD.

    :param url:  The link from which to extract the domain
    :type url: str
    :param include_subdomain: Whether or not to include the subdomain (e.g. 'news.google.com'); default is True
    :type include_subdomain: bool
    :param resolve_url: Whether to fully resolve the URL.  If False (default), it will operate on the URL as-is; if \
    True, the URL will be passed to :py:func:`pewtils.http.canonical_link` to be standardized prior to extracting the \
    domain.
    :param timeout: (Optional, for use with ``resolve_url``) Maximum number of seconds to wait on a request before \
    timing out (default is 1)
    :type timeout: int or float
    :param session: (Optional, for use with ``resolve_url``) A persistent session that can optionally be passed \
    (useful if you're processing many links at once)
    :type session: :py:class:`requests.Session` object
    :param user_agent: (Optional, for use with ``resolve_url``) User agent for the auto-created requests Session to use, \
    if a preconfigured requests Session is not provided
    :type user_agent: str
    :param expand_shorteners: If True, shortened URLs that don't successfully expand will be checked against a list \
    of known URL shorteners and expanded if recognized. (Default = True)
    :type expand_shorteners: bool
    :return: The domain for the link
    :rtype: str

    .. note:: If ``resolve_url`` is set to True, the link will be standardized prior to domain extraction (in which \
        case you can provide optional timeout, session, and user_agent parameters that will be passed to \
        :py:func:`pewtils.http.canonical_link`). By default, however, the link will be operated on as-is. The final \
        extracted domain is then checked against known URL shorteners (see :ref:`vanity_link_shorteners`) and if it \
        is recognized, the expanded domain will be returned instead. Shortened URLs that are not standardized and \
        do not follow patterns included in this dictionary of known shorteners may be returned with an incorrect domain.

    Usage::

        from pewtils.http import extract_domain_from_url

        >>> extract_domain_from_url("http://forums.bbc.co.uk", include_subdomain=False)
        "bbc.co.uk"
        >>> extract_domain_from_url("http://forums.bbc.co.uk", include_subdomain=True)
        "forums.bbc.co.uk"

    """

    if resolve_url:
        url = canonical_link(
            url, timeout=timeout, session=session, user_agent=user_agent
        )
    domain = tldextract.extract(url)
    if domain:
        if include_subdomain and domain.subdomain and domain.subdomain != "www":
            domain = ".".join([domain.subdomain, domain.domain, domain.suffix])
        else:
            domain = ".".join([domain.domain, domain.suffix])
        if expand_shorteners:
            domain = VANITY_LINK_SHORTENERS.get(domain, domain)
    return domain


def canonical_link(url, timeout=5.0, session=None, user_agent=None):

    """
    Tries to resolve a link to the "most correct" version.

    Useful for expanding short URLs from bit.ly / Twitter and for checking HTTP status codes without retrieving \
    the actual data. Follows redirects and tries to pick the most informative version of a URL while avoiding \
    redirects to generic 404 pages. Also tries to iteratively remove optional GET parameters.

    May not be particularly effective on dead links, but may still be able to follow redirects enough \
    to return a URL with the correct domain associated with the original link.

    :param url: The URL to test. Should be fully qualified.
    :type url: str
    :param timeout: How long to wait for a response before giving up (default is one second)
    :type timeout: int or float
    :param session: (Optional) A persistent session that can optionally be passed (useful if you're processing many \
    links at once)
    :type session: :py:class:`requests.Session` object
    :param user_agent: User agent for the auto-created requests Session to use, if a preconfigured requests Session \
    is not provided
    :type user_agent: str
    :return: The "canonical" URL as supplied by the server, or the original URL if none supplied.
    :rtype: str

    .. note:: See :ref:`link_shorteners` for a complete list of shortened links recognized by this function.

        This function might not resolve *all* existing URL modificiations, but it has been tested on a vast, well \
        maintained variety of URLs. It typically resolves URL to the correct final page while avoiding redirects to \
        generic error pages.

    Usage::

        from pewtils.http import canonical_link

        >>> canonical_link("https://pewrsr.ch/2lxB0EX")
        "https://www.pewresearch.org/interactives/how-does-a-computer-see-gender/"

    """

    BAD_STATUS_CODES = [
        302,
        307,
        400,
        404,
        405,
        407,
        500,
        501,
        502,
        503,
        504,
        520,
        530,
        404,
    ]
    PROXY_REQUIRED = [307, 407]
    CHECK_LENGTH = [301, 302, 200, 404]

    close_session = False
    if not session:
        close_session = True
        session = requests.Session()
        session.headers.update({"User-Agent": user_agent})
    if not url.startswith("http"):
        url = "http://" + url
    response = None
    try:
        with Timeout(timeout):
            try:
                response = session.head(url, allow_redirects=True, timeout=timeout)
            except requests.ConnectionError:
                try:
                    response = session.head(url, allow_redirects=False, timeout=timeout)
                except:
                    pass
    except:
        pass

    if response:

        history = [(h.status_code, h.url) for h in response.history]
        history.append((response.status_code, response.url))

        last_good_url = history[0][1]
        original_parsed = urlparse.urlparse(last_good_url)
        has_path = original_parsed.path not in ["/", ""]
        has_query = original_parsed.query != ""
        prev_was_shortener = False
        prev_path = None
        prev_query = None
        status_code = None
        for i, resp in enumerate(history):
            status_code, response_url = resp
            if "errors/404" in response_url:
                # If it's clearly a 404 landing page, stop and use the last observed good URL
                break
            parsed = urlparse.urlparse(response_url)
            if (
                parsed.netloc in VANITY_LINK_SHORTENERS.keys()
                or parsed.netloc in GENERAL_LINK_SHORTENERS
            ):
                # Don't consider known shortened URLs
                is_shortener = True
            else:
                is_shortener = False
            if not is_shortener:
                if i != 0:
                    for param, val in urlparse.parse_qs(parsed.query).items():
                        if len(val) == 1 and val[0].startswith("http"):
                            parsed_possible_url = urlparse.urlparse(val[0])
                            if (
                                parsed_possible_url.scheme
                                and parsed_possible_url.netloc
                            ):
                                # If the URL contains a GET parameter that is, itself, a URL, it's likely redirecting
                                # to it, so we're going to stop this run and start the process over with the new URL
                                return canonical_link(
                                    val[0],
                                    timeout=timeout,
                                    session=session,
                                    user_agent=user_agent,
                                )
                if status_code in PROXY_REQUIRED:
                    # These codes tend to indicate the last good URL in the chain
                    last_good_url = response_url
                    break
                good_path = not has_path or parsed.path not in ["/", ""]
                good_query = not has_query or parsed.query != ""
                # If the URL has a path or some GET parameters, we'll inspect further
                # Otherwise we just go with the previous URL
                # Link shorteners are very rarely used to reference root domains
                if good_query or good_path:
                    if (
                        re.sub("https", "http", response_url)
                        == re.sub("https", "http", last_good_url)
                        or parsed.path == original_parsed.path
                    ) or response_url.lower() == last_good_url.lower():
                        # If it's the same link but only the domain, protocol, or casing changed, it's fine
                        last_good_url = response_url
                    elif i != 0 and status_code in CHECK_LENGTH:
                        # For these codes, we're going to see how much the link changed
                        # Redirects and 404s sometimes preserve a decent URL, sometimes they go to a landing page
                        # The following cutoffs seem to do a good job most of the time:
                        # 1) The new URL has a long domain more than 7 characters, so it's not likely a shortened URL
                        # 2) The prior URL had a long path and this one has fewer than 20 characters and it wasn't
                        # swapped out for GET params
                        # 3) Or the prior URL had GET params and this one has far fewer and no replacement path
                        # If these conditions are met and the path or query do not identically match the prior link
                        # Then it's usually a generic error page
                        bad = False
                        if (
                            has_path
                            and len(parsed.netloc) > 7
                            and len(parsed.path) < 20
                            and len(parsed.query) == 0
                            and prev_path != parsed.path
                        ) or (
                            has_query
                            and len(parsed.netloc) > 7
                            and len(parsed.query) < 20
                            and len(parsed.path) <= 1
                            and prev_query != parsed.query
                        ):
                            bad = True
                        if not bad or prev_was_shortener:
                            last_good_url = response_url
                            # print("GOOD: {}, {}".format(status_code, response_url))
                        else:
                            # These can sometimes resolve further though, so we continue onward
                            prev_path = None
                            prev_query = None
                    else:
                        if status_code not in BAD_STATUS_CODES:
                            last_good_url = response_url
                        else:
                            break
                else:
                    # Resolved to a general URL
                    break

            prev_was_shortener = is_shortener
            prev_path = parsed.path
            prev_query = parsed.query

        if status_code not in BAD_STATUS_CODES:
            # If the URL ended on a good status code, we'll try to trim out any unnecessary GET parameters
            last_good_url = trim_get_parameters(
                last_good_url, session=session, timeout=timeout, user_agent=user_agent
            )

        url = last_good_url

    if close_session:
        session.close()

    return url



import re


URL_REGEX = re.compile(
    r"((?:https?:\/\/(?:www\.)?)?[-a-zA-Z0-9@:%._\+~#=]{1,4096}\.[a-z]{2,6}\b(?:[-a-zA-Z0-9@:%_\+.~#?&//=]*))"
)
"""
A compiled regular expression for extracting (probably) valid URLs.
"""

DOMAIN_REGEX = re.compile(
    r"(?:http[s]?\:\/\/)?(?:www(?:s?)\.)?([\w\.\-]+)(?:[\\\/](?:.+))?"
)
"""
A compiled regular expression for extracting domains from URLs. Can be useful in a pinch but we recommend \
using the :py:func:`pewtils.http.extract_domain_from_url` instead.
"""

HTTP_REGEX = re.compile(r"^http(?:s)?\:\/\/")
"""
A compiled regular expression for finding HTTP/S prefixes.
"""

US_DOLLAR_REGEX = re.compile(
    r"(\$(?:[1-9][0-9]{0,2}(?:(?:\,[0-9]{3})+)?(?:\.[0-9]{1,2})?))\b"
)
"""
A compiled regular expression finding USD monetary amounts.
"""

TITLEWORD_REGEX = re.compile(r"\b([A-Z][a-z]+)\b")
"""
A compiled regular expression for finding basic title-cased words.
"""

NUMBER_REGEX = re.compile(r"\b([0-9]+)\b")
"""
A compiled regular expression for finding raw numbers.
"""

NONALPHA_REGEX = re.compile(r"[^\w]")
"""
A compiled regular expression for finding non-alphanumeric values.
"""



from builtins import object
from contextlib import closing
from pewtils import is_not_null
from scandir import scandir
import boto3
import datetime
import hashlib
import json
import os
import pandas as pd
import pickle as pickle
import time

try:
    from io import StringIO, BytesIO

except ImportError:
    from StringIO import StringIO as BytesIO
    from StringIO import StringIO


class FileHandler(object):

    """
    Read/write data files in a variety of formats, locally and in Amazon S3 buckets.

    :param path: A valid path to the folder in local or s3 directory where files will be written to or read from
    :type path: str
    :param use_s3: Whether the path is an S3 location or local location
    :type use_s3: bool
    :param bucket: The name of the S3 bucket, required if ``use_s3=True``; will also try to fetch from the environment \
    as S3_BUCKET
    :type bucket: str

    .. note:: Typical rectangular data files (i.e. ``csv``, ``tab``, ``xlsx``, ``xls``, ``dta`` file extension types) will be \
        read to/written from a :py:class:`pandas.DataFrame` object. The exceptions are `pkl` and `json` objects which \
        accept any serializable Python object and correctly-formatted JSON object respectively.

    .. tip:: You can configure your environment to make it easier to automatically connect to S3 by defining the \
        variable ``S3_BUCKET``.

    Usage::

        from pewtils.io import FileHandler

        >>> h = FileHandler("./", use_s3=False)  # current local folder
        >>> df = h.read("my_csv", format="csv")
        # Do something and save to Excel
        >>> h.write("my_new_csv", df, format="xlsx")

        >>> my_data = [{"key": "value"}]
        >>> h.write("my_data", my_data, format="json")

        >>> my_data = ["a", "python", "list"]
        >>> h.write("my_data", my_data, format="pkl")

        # To read/write to an S3 bucket
        # The FileHandler detects your AWS tokens using boto3's standard methods to find them in ~/.aws or defined as environment variables.
        >>> h = FileHandler("/my_folder", use_s3=True, bucket="my-bucket")
    """

    def __init__(self, path, use_s3=None, bucket=None):
        self.bucket = os.environ.get("S3_BUCKET", None) if bucket is None else bucket
        self.path = path
        self.use_s3 = use_s3 if is_not_null(self.bucket) else False
        if self.use_s3:
            s3_params = {}
            self.s3 = boto3.client("s3")

        else:
            self.path = os.path.join(self.path)
            if not os.path.exists(self.path):
                try:
                    os.makedirs(self.path)

                except Exception as e:
                    print("Warning: couldn't make directory '{}'".format(self.path))
                    print(e)

    def iterate_path(self):

        """
        Iterates over the directory and returns a list of filenames or S3 object keys

        :return: Yields a list of filenames or S3 keys
        :rtype: iterable

        Usage::

            from pewtils.io import FileHandler

            >>> h = FileHandler("./", use_s3=False)
            >>> for file in h.iterate_path(): print(file)
            file1.csv
            file2.pkl
            file3.json

        """

        if self.use_s3:
            for key in self.s3.list_objects(Bucket=self.bucket, Prefix=self.path)['Contents']:
                yield key["Key"]

        else:
            for f in scandir(self.path):
                yield f.name

    def clear_folder(self):
        """
        Deletes the path (if local) or unlinks all keys in the bucket folder (if S3)

        .. warning:: This is a destructive function, use with caution!

        Usage::

            from pewtils.io import FileHandler

            >>> h = FileHandler("./", use_s3=False)
            >>> len(list(h.iterate_path()))
            3
            >>> h.clear_folder()
            >>> len(list(h.iterate_path()))
            0

        """

        if self.use_s3:
            for key in self.s3.list_objects(Bucket=self.bucket, Prefix=self.path)['Contents']:
                self.s3.delete_object(Bucket=self.bucket, Prefix=key['Key'])

        else:
            for f in scandir(self.path):
                os.unlink(os.path.join(self.path, f.name))

    def clear_file(self, key, format="pkl", hash_key=False):
        """
        Deletes a specific file.

        .. warning:: This is a destructive function, use with caution!

        :param key: The name of the file to delete
        :type key: str
        :param format: The file extension
        :type format: str
        :param hash_key: If True, will hash the filename before looking it up; default is False.
        :type hash_key: bool

        Usage::

            from pewtils.io import FileHandler

            >>> h = FileHandler("./", use_s3=False)
            >>> for file in h.iterate_path(): print(file)
            file1.csv
            file2.pkl
            file3.json
            >>> h.clear_file("file1", format="csv")
            >>> for file in h.iterate_path(): print(file)
            file2.pkl
            file3.json

        """

        if hash_key:
            key = self.get_key_hash(key)

        if self.use_s3:
            filepath = "/".join([self.path, "{}.{}".format(key, format)])
            key = self.s3.delete_object(Bucket=self.bucket, Key=filepath)

        else:
            key += ".{}".format(format)
            path = os.path.join(self.path, key)
            os.unlink(path)

    def get_key_hash(self, key):

        """
        Converts a key to a hashed representation. Allows you to pass arbitrary objects and convert their string \
        representation into a shorter hashed key, so it can be useful for caching. You can call this method \
        directly to see the hash that a key will be converted into, but this method is mainly used in conjunction \
        with the :py:meth:`pewtils.FileHandler.write` and :py:meth:`pewtils.FileHandler.read` methods by passing in \
        ``hash_key=True``.

        :param key: A raw string or Python object that can be meaningfully converted into a string representation
        :type key: str or object
        :return: A SHA224 hash representation of that key
        :rtype: str

        Usage::

            from pewtils.io import FileHandler

            >>> h = FileHandler("tests/files", use_s3=False)
            >>> h.get_key_hash("temp")
            "c51bf90ccb22befa316b7a561fe9d5fd9650180b14421fc6d71bcd57"
            >>> h.get_key_hash({"key": "value"})
            "37e13e1116c86a6e9f3f8926375c7cb977ca74d2d598572ced03cd09"

        """

        try:
            return hashlib.sha224(key.encode("utf8")).hexdigest()
        except AttributeError:
            return hashlib.sha224(str(key).encode("utf8")).hexdigest()

    def write(
        self, key, data, format="pkl", hash_key=False, add_timestamp=False, **io_kwargs
    ):

        """
        Writes arbitrary data objects to a variety of file formats.


        :param key: The name of the file or key (without a file suffix!)
        :type key: str
        :param data: The actual data to write to the file
        :type data: object
        :param format: The format the data should be saved in (pkl/csv/tab/xlsx/xls/dta/json). Defaults to pkl. \
        This will be used as the file's suffix.
        :type format: str
        :param hash_key: Whether or not to hash the provided key before saving the file. (Default=False)
        :type hash_key: bool
        :param add_timestamp: Optionally add a timestamp to the filename
        :type add_timestamp: bool
        :param io_kwargs: Additional parameters to pass along to the Pandas save function, if applicable
        :return: None

        .. note:: When saving a ``csv``, ``tab``, ``xlsx``, ``xls``, or ``dta`` file, this function expects to receive a \
            Pandas :py:class:`pandas.DataFrame`. When you use these formats, you can also pass optional ``io_kwargs`` \
            which will be forwarded to the corresponding :py:mod:`pandas` method below:

                - `dta`: :py:meth:`pandas.DataFrame.to_stata`
                - `csv`: :py:meth:`pandas.DataFrame.to_csv`
                - `tab`: :py:meth:`pandas.DataFrame.to_csv`
                - `xlsx`: :py:meth:`pandas.DataFrame.to_excel`
                - `xls`: :py:meth:`pandas.DataFrame.to_excel`

            If you're trying to save an object to JSON, it assumes that you're passing it valid JSON. By default, \
            the handler attempts to use pickling, allowing you to save anything you want, as long as it's serializable.

        """

        format = format.strip(".")

        if hash_key:
            key = self.get_key_hash(key)

        if add_timestamp:
            key = "{}_{}".format(key, datetime.datetime.now())

        def _get_output(output, data, io_kwargs):
            if format == "tab":
                io_kwargs["sep"] = "\t"
            if format in ["csv", "tab"]:
                data.to_csv(output, encoding="utf8", **io_kwargs)
            elif format == "dta":
                data.to_stata(output, **io_kwargs)
            elif format in ["xls", "xlsx"]:
                writer = pd.ExcelWriter(output, engine="xlsxwriter")
                data.to_excel(writer, **io_kwargs)
                writer.save()
            data = output.getvalue()
            return data

        if format in ["csv", "xls", "xlsx", "tab", "dta"]:
            try:
                data = _get_output(BytesIO(), data, io_kwargs)
            except Exception as e:
                try:
                    data = _get_output(StringIO(), data, io_kwargs)
                except:
                    raise Exception(
                        "Couldn't convert data into '{}' format".format(format)
                    )

        elif format == "pkl":
            data = pickle.dumps(data, **io_kwargs)
        elif format == "json":
            data = json.dumps(data, **io_kwargs)

        key += ".{}".format(format)

        if self.use_s3:
            try:
                upload = BytesIO(data)

            except TypeError:
                upload = BytesIO(data.encode())

            self.s3.upload_fileobj(upload, Bucket=self.bucket, Key="/".join([self.path, key]))

        else:
            path = os.path.join(self.path, key)
            if os.path.exists(self.path):
                try:
                    with closing(open(path, "w")) as output:
                        output.write(data)
                except:
                    with closing(open(path, "wb")) as output:
                        output.write(data)

    def read(self, key, format="pkl", hash_key=False, **io_kwargs):

        """
        Reads a file from the directory or S3 path, returning its contents.

        :param key: The name of the file to read (without a suffix!)
        :type key: str
        :param format: The format of the file (pkl/json/csv/dta/xls/xlsx/tab); expects the file extension to match
        :type format: str
        :param hash_key: Whether the key should be hashed prior to looking for and retrieving the file.
        :type hash_key: bool
        :param io_kwargs: Optional arguments to be passed to the specific load function (dependent on file format)
        :return: The file contents, in the requested format

        .. note:: You can pass optional ``io_kwargs`` that will be forwarded to the function below that corresponds to \
            the format of the file you're trying to read in

            - `dta`: :py:meth:`pandas.DataFrame.read_stata`
            - `csv`: :py:meth:`pandas.DataFrame.read_csv`
            - `tab`: :py:meth:`pandas.DataFrame.read_csv`
            - `xlsx`: :py:meth:`pandas.DataFrame.read_excel`
            - `xls`: :py:meth:`pandas.DataFrame.read_excel`
        """

        format = format.strip(".")

        if hash_key:
            key = self.get_key_hash(key)

        data = None
        filepath = "/".join([self.path, "{}.{}".format(key, format)])

        if self.use_s3:
            try:
                data = StringIO()

            except TypeError:
                data = BytesIO()

            self.s3.download_fileobj(data, Bucket=self.bucket, Key=filepath)
            data = data.getvalue()
        else:
            if os.path.exists(filepath):
                try:
                    with closing(open(filepath, "r")) as infile:
                        data = infile.read()

                except:
                    # TODO: handle this exception more explicitly
                    with closing(open(filepath, "rb")) as infile:
                        data = infile.read()

        if is_not_null(data):
            if format == "pkl":
                try:
                    data = pickle.loads(data)

                except TypeError:
                    data = None

                except ValueError:
                    if "attempt_count" not in io_kwargs:
                        io_kwargs["attempt_count"] = 1

                    print(
                        "Insecure pickle string; probably a concurrent read-write, \
                        will try again in 5 seconds (attempt #{})".format(
                            io_kwargs["attempt_count"]
                        )
                    )
                    time.sleep(5)

                    if io_kwargs["attempt_count"] <= 3:
                        io_kwargs["attempt_count"] += 1
                        data = self.read(
                            key, format=format, hash_key=hash_key, **io_kwargs
                        )

                    else:
                        data = None

                except Exception as e:
                    print("Couldn't load pickle!  {}".format(e))
                    data = None

            elif format in ["tab", "csv"]:
                if format == "tab":
                    io_kwargs["delimiter"] = "\t"

                try:
                    data = pd.read_csv(BytesIO(data), **io_kwargs)

                except:
                    data = pd.read_csv(StringIO(data), **io_kwargs)

            elif format in ["xlsx", "xls"]:
                # https://stackoverflow.com/questions/64264563/attributeerror-elementtree-object-has-no-attribute-getiterator-when-trying
                if "engine" not in io_kwargs:
                    io_kwargs["engine"] = "openpyxl"

                try:
                    data = pd.read_excel(BytesIO(data), **io_kwargs)

                except:
                    data = pd.read_excel(StringIO(data), **io_kwargs)

            elif format == "json":
                try:
                    data = json.loads(data)

                except:
                    pass

            elif format == "dta":
                try:
                    data = pd.read_stata(BytesIO(data), **io_kwargs)

                except:
                    data = pd.read_stata(StringIO(data), **io_kwargs)

            elif format == "txt":
                if isinstance(data, bytes):
                    data = data.decode()

        return data



from setuptools import setup, find_packages

with open("README.md") as README:
    readme = str(README.read())

with open("requirements.txt") as reqs:
    lines = reqs.read().split("\n")
    install_requires = [line for line in lines if line]

setup(
    name="pewtils",
    version="1.1.5",
    description="General programming utilities from Pew Research Center",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/pewresearch/pewtils",
    author="Pew Research Center",
    author_email="info@pewresearch.org",
    install_requires=install_requires,
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    include_package_data=True,
    keywords="utilities, link standardization, input, output",
    license="GPLv2+",
    classifiers=[
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        # "Development Status :: 6 - Mature",
        # "Development Status :: 7 - Inactive",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
)
